@@ -7,9 +7,11 @@ use nix;
 use nix::errno::Errno;
 use nix::sys::select::{select, FdSet};
 use nix::sys::signal::{signal, SigHandler, Signal};
+use nix::sys::time::{TimeVal, TimeValLike};
 use std::fs::File;
 use std::io::prelude::*;
 use std::sync::atomic::{AtomicBool, Ordering};
+use std::time::{Duration, SystemTime};
 
 static RUNNING: AtomicBool = AtomicBool::new(true);
 
@@ -72,8 +74,11 @@ impl App {
         let win_fd = self.win.fd();
         let pty_fd = self.term.pty.fd();
         let mut buf = [0; 8192];
+        let mut last_blink = SystemTime::now();
+        let blink_duration = Duration::from_millis(500);
 
         while is_running() {
+            let blink_elapsed = last_blink.elapsed().map_or_else(|_| blink_duration, |e| e);
             let mut rfds = FdSet::new();
             rfds.insert(pty_fd);
             rfds.insert(win_fd);
@@ -83,7 +88,17 @@ impl App {
                 wfds.insert(pty_fd);
             }
 
-            match select(None, Some(&mut rfds), Some(&mut wfds), None, None) {
+            // Something pending so let select just return otherwise events
+            // might be delayed.
+            let mut timeout_in = if self.win.is_pending() {
+                TimeVal::milliseconds(0)
+            } else if let Some(to) = blink_duration.checked_sub(blink_elapsed) {
+                TimeVal::milliseconds(to.as_millis() as i64)
+            } else {
+                TimeVal::milliseconds(blink_duration.as_millis() as i64)
+            };
+            let timeout = Some(&mut timeout_in);
+            match select(None, Some(&mut rfds), Some(&mut wfds), None, timeout) {
                 Ok(_) => (),
                 Err(Errno::EINTR) => continue,
                 Err(err) => return Err(err.into()),
@@ -95,16 +110,19 @@ impl App {
 
             self.win.undraw_cursor(&self.term);
 
-            if rfds.contains(win_fd) {
-                self.win.process_input(&mut self.term);
-            }
+            // Let pending do it's thing so always try to process events.
+            self.win.process_input(&mut self.term);
+
             if rfds.contains(pty_fd) {
                 let n = self.term.pty.read(&mut buf)?;
                 self.log_pty(&buf[..n])?;
                 self.vte
                     .process_input(&buf[..n], &mut self.win, &mut self.term);
             }
-
+            if blink_elapsed >= blink_duration {
+                self.win.toggle_blink();
+                last_blink = SystemTime::now();
+            }
             self.win.draw(&mut self.term);
         }
 
