@@ -1,8 +1,9 @@
 // FIXME: support wide chars
 
 use crate::charset::CharsetTable;
+use crate::color::{BG_COLOR, CURSOR_COLOR, CURSOR_REV_COLOR, FG_COLOR};
 use crate::cursor::Cursor;
-use crate::glyph::{blank_glyph, Glyph, GlyphAttr};
+use crate::glyph::{blank_glyph, Glyph, GlyphAttr, GlyphProp};
 use crate::point::Point;
 use crate::pty::Pty;
 use crate::snap::{is_delim, SnapMode};
@@ -59,6 +60,7 @@ pub struct Term {
     pub scroll_top: usize,
     pub scroll_bot: usize,
     pub charset: CharsetTable,
+    pub prop: GlyphProp,
     saved_c: Option<Cursor>,
     saved_alt_c: Option<Cursor>,
     // lines contains the main and alternate screens, access it through the
@@ -81,6 +83,7 @@ impl Term {
             scroll_top: 0,
             scroll_bot: 0,
             charset: CharsetTable::new(),
+            prop: GlyphProp::new(FG_COLOR, BG_COLOR, GlyphAttr::empty()),
             mode: TermMode::WRAP,
             tabs: Vec::new(),
             sel: Selection::new(),
@@ -100,6 +103,10 @@ impl Term {
             return false;
         }
 
+        if !self.sel.empty {
+            self.clear_selection();
+        }
+
         if self.c.y > rows - 1 {
             self.scroll_up(0, self.c.y - rows + 1)
         }
@@ -151,12 +158,15 @@ impl Term {
         g
     }
 
-    pub fn get_glyph_at_cursor(&self) -> (usize, usize, Glyph) {
+    pub fn get_glyph_at_cursor(&self) -> Glyph {
         let (x, y) = (self.c.x, self.c.y);
         let mut g = self.lines()[y][x];
-        let reverse = !self.is_selected(x, y);
-        g.prop = self.c.glyph.prop.resolve(reverse);
-        (x, y, g)
+        if self.is_selected(x, y) {
+            g.prop.bg = CURSOR_REV_COLOR;
+        } else {
+            g.prop.bg = CURSOR_COLOR;
+        }
+        g
     }
 
     pub fn reset(&mut self) {
@@ -190,10 +200,11 @@ impl Term {
         R1: Iterator<Item = usize> + Clone,
         R2: Iterator<Item = usize>,
     {
+        let mut glyph = blank_glyph();
+        glyph.prop = self.prop;
         for y in yrange {
             self.dirty[y] = true;
             for x in xrange.clone() {
-                let glyph = self.c.glyph;
                 self.lines_mut()[y][x].clear(glyph);
                 if self.is_selected(x, y) {
                     self.clear_selection();
@@ -354,7 +365,7 @@ impl Term {
         // x, y may have updated.
         let (x, y) = (self.c.x, self.c.y);
         self.dirty[y] = true;
-        self.lines_mut()[y][x].prop = self.c.glyph.prop;
+        self.lines_mut()[y][x].prop = self.prop;
         self.lines_mut()[y][x].c = c;
         for x2 in x + 1..x + width {
             self.lines_mut()[y][x2].prop.attr.insert(GlyphAttr::DUMMY);
@@ -414,9 +425,9 @@ impl Term {
     }
 
     pub fn start_selection(&mut self, x: usize, y: usize, mode: SnapMode) {
-        // clear previous selection
-        self.clear_selection();
-        self.set_dirty(self.sel.nb.y..=self.sel.ne.y);
+        if !self.sel.empty {
+            self.clear_selection();
+        }
 
         self.sel.mode = mode;
         self.sel.ob.x = cmp::min(x, self.cols - 1);
@@ -431,6 +442,7 @@ impl Term {
     }
 
     pub fn extend_selection(&mut self, x: usize, y: usize) {
+        self.clear_selection();
         self.sel.oe.x = cmp::min(x, self.cols - 1);
         self.sel.oe.y = cmp::min(y, self.rows - 1);
         self.normalize_selection();
@@ -445,6 +457,7 @@ impl Term {
 
     pub fn clear_selection(&mut self) {
         self.sel.empty = true;
+        self.set_dirty(self.sel.nb.y..=self.sel.ne.y);
     }
 
     pub fn get_selection_content(&self) -> Option<String> {
@@ -481,6 +494,18 @@ impl Term {
         self.set_dirty(0..self.dirty.len());
     }
 
+    pub fn setdirtattr(&mut self, attr: GlyphAttr) {
+        let start = self.first_line();
+        for i in 0..self.rows {
+            for g in self.lines[start + i].iter() {
+                if g.prop.attr.contains(attr) {
+                    self.dirty[i] = true;
+                    break;
+                }
+            }
+        }
+    }
+
     fn normalize_selection(&mut self) {
         let (mut nb, mut ne) = sort_pair(self.sel.ob, self.sel.oe);
 
@@ -510,10 +535,7 @@ impl Term {
             }
         }
 
-        let bot = cmp::min(nb.y, self.sel.nb.y);
-        let top = cmp::max(ne.y, self.sel.ne.y);
-        self.set_dirty(bot..=top);
-
+        self.set_dirty(nb.y..=ne.y);
         self.sel.empty = false;
         self.sel.nb = nb;
         self.sel.ne = ne;
@@ -604,6 +626,15 @@ impl Term {
             .contains(GlyphAttr::WRAP)
     }
 
+    #[inline]
+    fn first_line(&self) -> usize {
+        if self.mode.contains(TermMode::ALTSCREEN) {
+            self.rows
+        } else {
+            0
+        }
+    }
+
     #[inline]
     fn lines(&self) -> &[Vec<Glyph>] {
         if self.mode.contains(TermMode::ALTSCREEN) {
