@@ -1,5 +1,9 @@
 use crate::app::app_exit;
-use crate::color::{BG_COLOR, BG_COLOR_NAME, FG_COLOR_NAME};
+use crate::color::{
+    BG_COLOR, BG_COLOR_NAME, CURSOR_COLOR, CURSOR_COLOR_NAME, CURSOR_REV_COLOR,
+    CURSOR_REV_COLOR_NAME, FG_COLOR_NAME,
+};
+use crate::cursor::CursorMode;
 use crate::font::Font;
 use crate::glyph::{GlyphAttr, GlyphProp};
 use crate::keymap::map_key;
@@ -38,11 +42,17 @@ bitflags! {
 // FIXME, this can only be 0 until impemented everywhere.
 const BORDERPX: u16 = 0;
 
+/*
+ * thickness of underline and bar cursors
+ */
+const CURSORTHICKNESS: usize = 2;
+
 const FORCEMOUSEMOD: u32 = x11::ShiftMask;
 
 pub struct Win {
     visible: bool,
     mode: WinMode,
+    cursor_vis: bool,
 
     dpy: x11::Display,
     win: x11::Window,
@@ -94,7 +104,7 @@ impl Win {
         let (width, height) = (cols * cw, rows * ch);
 
         let cmap = x11::XDefaultColormap(dpy, scr);
-        let mut colors = Vec::with_capacity(257); //COLOR_NAMES.len());
+        let mut colors = Vec::with_capacity(260);
         for i in 0..=255 {
             colors.push(
                 x11::xloadcolor(dpy, vis, cmap, i, None).expect("Failed to load a default color!"),
@@ -102,12 +112,12 @@ impl Win {
         }
         // cursor
         colors.push(
-            x11::xloadcolor(dpy, vis, cmap, 256, Some("#cccccc"))
+            x11::xloadcolor(dpy, vis, cmap, 256, Some(CURSOR_COLOR_NAME))
                 .expect("Failed to load a default color!"),
         );
         // reverse cursor
         colors.push(
-            x11::xloadcolor(dpy, vis, cmap, 257, Some("#555555"))
+            x11::xloadcolor(dpy, vis, cmap, 257, Some(CURSOR_REV_COLOR_NAME))
                 .expect("Failed to load a default color!"),
         );
         // foreground
@@ -185,6 +195,7 @@ impl Win {
         Ok(Win {
             visible: true,
             mode: WinMode::empty(),
+            cursor_vis: true,
 
             sel_type,
             sel_snap: Snap::new(),
@@ -276,11 +287,19 @@ impl Win {
             }
         }
         // cursor
-        if let Ok(color) = x11::xloadcolor(self.dpy, self.vis, self.cmap, 256, Some("#cccccc")) {
+        if let Ok(color) =
+            x11::xloadcolor(self.dpy, self.vis, self.cmap, 256, Some(CURSOR_COLOR_NAME))
+        {
             self.colors.push(color);
         }
         // reverse cursor
-        if let Ok(color) = x11::xloadcolor(self.dpy, self.vis, self.cmap, 257, Some("#555555")) {
+        if let Ok(color) = x11::xloadcolor(
+            self.dpy,
+            self.vis,
+            self.cmap,
+            257,
+            Some(CURSOR_REV_COLOR_NAME),
+        ) {
             self.colors.push(color);
         }
         // foreground
@@ -319,11 +338,17 @@ impl Win {
         self.draw_cells(&[g.c], g.prop, x * self.cw, y * self.ch);
     }
 
+    pub fn toggle_blink(&mut self) {
+        self.mode
+            .set(WinMode::BLINK, !self.mode.contains(WinMode::BLINK));
+    }
+
     pub fn draw(&mut self, term: &mut Term) {
         if !self.visible {
             return;
         }
 
+        term.setdirtattr(GlyphAttr::BLINK);
         for y in 0..term.rows {
             if !term.dirty[y] {
                 continue;
@@ -332,9 +357,49 @@ impl Win {
             term.dirty[y] = false;
         }
 
-        if !self.mode.contains(WinMode::HIDE) {
-            let (x, y, g) = term.get_glyph_at_cursor();
-            self.draw_cells(&[g.c], g.prop, x * self.cw, y * self.ch);
+        if term.c.blink {
+            self.cursor_vis = !self.mode.contains(WinMode::BLINK);
+        } else {
+            self.cursor_vis = true;
+        }
+        if !self.mode.contains(WinMode::HIDE) && self.cursor_vis {
+            let (x, y) = (term.c.x, term.c.y);
+            match term.c.mode {
+                CursorMode::Block => {
+                    let g = term.get_glyph_at_cursor();
+                    self.draw_cells(&[g.c], g.prop, x * self.cw, y * self.ch);
+                }
+                CursorMode::Underline => {
+                    let drawcol = if term.is_selected(x, y) {
+                        self.colors[CURSOR_REV_COLOR]
+                    } else {
+                        self.colors[CURSOR_COLOR]
+                    };
+                    x11::XftDrawRect(
+                        self.draw,
+                        &drawcol,
+                        BORDERPX as usize + x * self.cw,
+                        BORDERPX as usize + (y + 1) * self.ch - CURSORTHICKNESS,
+                        self.cw,
+                        CURSORTHICKNESS,
+                    );
+                }
+                CursorMode::Bar => {
+                    let drawcol = if term.is_selected(x, y) {
+                        self.colors[CURSOR_REV_COLOR]
+                    } else {
+                        self.colors[CURSOR_COLOR]
+                    };
+                    x11::XftDrawRect(
+                        self.draw,
+                        &drawcol,
+                        BORDERPX as usize + x * self.cw,
+                        BORDERPX as usize + y * self.ch,
+                        CURSORTHICKNESS,
+                        self.ch,
+                    );
+                }
+            }
         }
 
         self.finish_draw(term.cols, term.rows);
@@ -347,6 +412,10 @@ impl Win {
         self.draw(term);
     }
 
+    pub fn is_pending(&self) -> bool {
+        x11::XPending(self.dpy) > 0
+    }
+
     pub fn process_input(&mut self, term: &mut Term) {
         while x11::XPending(self.dpy) > 0 {
             let mut xev = x11::XNextEvent(self.dpy);
@@ -710,8 +779,13 @@ impl Win {
         if attr.contains(GlyphAttr::BOLD) && fg < 8 {
             fg += 8;
         }
-
-        let fg = if fg & (1 << 24) > 0 {
+        let charlen = if attr.contains(GlyphAttr::WIDE) {
+            cs.len() * 2
+        } else {
+            cs.len()
+        };
+        let width = charlen * self.cw;
+        let mut fg = if fg & (1 << 24) > 0 {
             // truecolor
             self.to_truecolor(fg)
         } else {
@@ -725,12 +799,42 @@ impl Win {
         };
         let font = self.font.get(attr);
 
-        x11::XftDrawRect(self.draw, &bg, xp, yp, cs.len() * self.cw, self.ch);
+        if attr.contains(GlyphAttr::FAINT) {
+            let faintfg = x11::XRenderColor {
+                alpha: fg.color.alpha,
+                red: fg.color.red / 2,
+                green: fg.color.green / 2,
+                blue: fg.color.blue / 2,
+            };
+            if let Ok(nfg) = x11::XftColorAllocValue(self.dpy, self.vis, self.cmap, &faintfg) {
+                fg = nfg;
+            } else {
+                println!("Failed to alloc truecolor for FAINT")
+            }
+        }
+        if attr.contains(GlyphAttr::INVISIBLE) {
+            fg = bg;
+        }
+        if attr.contains(GlyphAttr::BLINK) && self.mode.contains(WinMode::BLINK) {
+            fg = bg;
+        }
+
+        x11::XftDrawRect(self.draw, &bg, xp, yp, width, self.ch);
         let idx = cs
             .iter()
             .map(|&c| x11::XftCharIndex(self.dpy, font, c))
             .collect::<Vec<u32>>();
         x11::XftDrawGlyphs(self.draw, &fg, font, xp, yp + self.ca, &idx);
+
+        /* Render underline and strikethrough. */
+        if attr.contains(GlyphAttr::UNDERLINE) {
+            let y = yp + self.font.ascent() + 1;
+            x11::XftDrawRect(self.draw, &fg, xp, y, width, 1);
+        }
+        if attr.contains(GlyphAttr::STRUCK) {
+            let y = yp + (2 * self.font.ascent() / 3);
+            x11::XftDrawRect(self.draw, &fg, xp, y, width, 1);
+        }
     }
 
     fn draw_line(&mut self, term: &mut Term, y: usize) {
